/*
** Module   : PropertyDefinition.java
** Abstract : A container for a DMO property's metadata.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20130618 Created initial version.
*/

package com.goldencode.p2j.persist;

import java.io.*;

import org.hibernate.usertype.UserType;

/**
 * Container for a DMO property's metadata: name, type and extent.  Used to send the metadata of a
 * DMO to a remote side, when remote appserver calls are involved.
 */
public class PropertyDefinition
implements Externalizable
{
   /** Default value for no-extent properties. */
   private static final int NO_EXTENT = -2;
   
   /**
    * The P2J-style type of this property, the name of one of P2J's {@link UserType} sub-classes.
    */
   private String type;

   /** The extent of this property, or {@link #NO_EXTENT} if the property has no extent. */
   private int extent;
   
   /** The name of this property. */
   private String name;
      
   /**
    * Default c'tor, explicitly added to allow instances of this class to be created on 
    * deserialization. Not for public use.
    */
   public PropertyDefinition()
   {
   }
   
   /**
    * Create a new, no-extent, property definition.
    * 
    * @param    name
    *           The name of this property.
    * @param    type
    *           The type of this property.
    */
   public PropertyDefinition(String name, String type)
   {
      this.name = name;
      this.type = type;
      this.extent = NO_EXTENT;
   }
   
   /**
    * Create a new property definition.
    * 
    * @param    name
    *           The name of this property.
    * @param    type
    *           The type of this property.
    * @param    extent
    *           The extent of this property.
    */
   public PropertyDefinition(String name, String type, int extent)
   {
      this.name = name;
      this.type = type;
      this.extent = extent;
   }

   /**
    * Get the type of this property.
    * 
    * @return   See above.
    */
   public String getType()
   {
      return type;
   }

   /**
    * Get the extent of this property.
    * 
    * @return   See above.
    */
   public int getExtent()
   {
      return extent;
   }

   /**
    * Get the name of this property.
    * 
    * @return   See above.
    */
   public String getName()
   {
      return name;
   }
   
   /**
    * Check if this property is an extent property.
    * 
    * @return   <code>true</code> if this property is an extent property.
    */
   public boolean isExtent()
   {
      return extent != NO_EXTENT;
   }

   /**
    * Send the property definition to the specified output destination.
    * 
    * @param    out
    *           The output destination to which the property definition will be sent.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   public final void writeExternal(ObjectOutput out)
   throws IOException
   {
      out.writeUTF(name);
      out.writeUTF(type);
      out.writeInt(extent);
   }

   /**
    * Read the property definition from the specified input source.
    * 
    * @param    in
    *           The input source from which the property definition will be read.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   public final void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      name = in.readUTF();
      type = in.readUTF();
      extent = in.readInt();
   }
}
