/*
** Module   : AgentPool.java
** Abstract : Definition and implementation of agent pools (based on appserver's operating mode).
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20130529 Created initial version.
*/

package com.goldencode.p2j.util;

import java.util.*;

import org.apache.commons.logging.*;
import com.goldencode.p2j.*;
import com.goldencode.p2j.directory.*;
import com.goldencode.p2j.main.*;
import com.goldencode.p2j.security.SecurityManager;
import com.goldencode.p2j.util.Agent.InvocationModes;

/**
 * The main class for agent pooling.  Based on the appserver's operating mode, the 
 * {@link #newConnection()} call will return a:
 * <ul>
 * <li> {@link BoundPool} for State-reset and State-aware modes.</li>
 * <li> {@link UnboundPool} for State-free and Stateless modes.</li>
 * </ul>
 * Agents are extracted from {@link #mainPool} as connection or procedure requests are incoming
 * and put back in the pool only after the connection terminates or it becomes unbound.
 */
abstract class AgentPool
{
   /** Logger. */
   private static final Log LOG = LogFactory.getLog(AgentPool.class);

   /** Map of connection IDs to appserver name. */
   private static final Map<String, String> idsToAppserver = new HashMap<String, String>();

   // TODO: limit each appserver to be started by only one account
   /** Map of subject IDs to appserver names. */
   private static final Map<String, String> subjectToAppserver = new HashMap<String, String>();

   /** The next connection ID; starts from a random value. */
   private static long nextId = new Random().nextLong();

   /** The pool of agents.*/
   private final Set<Agent> mainPool = new HashSet<Agent>();
   
   /** All started agents. */
   private final Set<Agent> all = new HashSet<Agent>();
   
   /** The remote persistent procedures exposed as proxy procedures to the requester. */
   private final Map<String, Map<String, PersistentProcedure>>
      persistentProcedures = new HashMap<String, Map<String, PersistentProcedure>>();
   
   /** The agent manager. */
   private final Agent manager;

   /**
    * This instance is used by the manager to send calls to the P2J client, so agents can be
    * started.
    */
   private final AppServer handler;

   /** The appserver definition. */
   private final AppServerDefinition appDef;
   
   /**
    * Build a new agent pool, for the specified appserver.
    * 
    * @param    def
    *           The appserver definition.
    * @param    handler
    *           The handler used to start agents.
    */
   private AgentPool(AppServerDefinition def, AppServer handler)
   {
      this.handler = handler;
      this.appDef = def;
      this.manager = new Agent(this, this.appDef.propath);
   }
   
   /**
    * Get the appserver name started by the given subject ID.
    * 
    * @param    subjectId
    *           The subject ID used to start an appserver.
    *           
    * @return   See above.
    */
   static String getAppserverBySubject(String subjectId)
   {
      synchronized (subjectToAppserver)
      {
         return subjectToAppserver.get(subjectId);
      }
   }
   
   /**
    * Get the appserver name used to established the given connection.
    * 
    * @param    connectionId
    *           The connection ID.
    *           
    * @return   See above.
    */
   static String getAppserver(String connectionId)
   {
      synchronized (idsToAppserver)
      {
         return idsToAppserver.get(connectionId);
      }
   }
   
   /**
    * Instantiate a new agent pool for the given appserver, based on the operating mode.
    * 
    * @param    appServer
    *           The appserver's name.
    * @param    handler
    *           The handler used to start agents.
    *           
    * @return   A new pool instance.
    * 
    * @throws   IllegalArgumentException
    *           If the operating mode can not be resolved to one of {@link AppServerOperatingMode}
    *           constants.
    */
   static AgentPool newPool(String appServer, AppServer handler)
   {
      AppServerDefinition appDef = new AppServerDefinition(appServer);
      switch (appDef.operatingMode)
      {
         case STATE_AWARE:
            return new BoundPool(appDef, handler, false);
         case STATE_RESET:
            return new BoundPool(appDef, handler, true);
         case STATELESS:
            return new UnboundPool(appDef, handler, false);
         case STATE_FREE:
            return new UnboundPool(appDef, handler, true);
      }
   
      throw new IllegalArgumentException(
         "Illegal operating mode: " + appDef.operatingMode + ".");
   }

   /**
    * Get the next connection ID.
    * 
    * @return   See above.
    */
   protected static synchronized String nextId()
   {
      do
      {
         String next = Long.toString(++nextId);
         synchronized (idsToAppserver)
         {
            if (!idsToAppserver.containsKey(next))
               return next;
         }
      }
      while (true);
   }

   /**
    * Acquire a new agent from the pool to handle requests incoming via the given connection ID.
    * 
    * @param    id
    *           The connection ID.
    *           
    * @return   See above.
    */
   abstract Agent acquireAgent(String id);

   /**
    * Release the given agent and place it back in the pool. This will handle any bound-related
    * constraints.
    * 
    * @param    id
    *           The connection ID.
    * @param    agent
    *           The agent which needs to be released.
    */
   abstract void releaseAgent(String id, Agent agent);

   /**
    * Establish a new connection.
    * 
    * @return   The connection ID.
    * 
    * @throws   NumberedException
    *           If the connection could not be established.
    */
   abstract String newConnection()
   throws NumberedException;

   /**
    * Get the name of the appserver handled by this pool.
    * 
    * @return   See above.
    */
   String getName()
   {
      return appDef.appServer;
   }

   /**
    * Get the startup procedure for the appserver handled by this pool.
    * 
    * @return   See above.
    */
   String getStartup()
   {
      return appDef.startup;
   }

   /**
    * Get the startup parameters for the appserver handled by this pool.
    * 
    * @return   See above.
    */
   String getStartupParameter()
   {
      return appDef.startupParameter;
   }

   /**
    * Get the shutdown procedure for the appserver handled by this pool.
    * 
    * @return   See above.
    */
   String getShutdown()
   {
      return appDef.shutdown;
   }

   /**
    * Get the activate procedure for the appserver handled by this pool.
    * 
    * @return   See above.
    */
   String getActivate()
   {
      return appDef.activate;
   }

   /**
    * Get the deactivate procedure for the appserver handled by this pool.
    * 
    * @return   See above.
    */
   String getDeactivate()
   {
      return appDef.deactivate;
   }

   /**
    * Establish a new appserver connection. This will invoke any connect procedure for this 
    * appserver, to which the given <code>user</code>, <code>pwd</code> and <code>serverInfo</code>
    * parameters will be passed.
    * <p>
    * The appserver connection must be configured to start under the given P2J-level subject ID.
    *  
    * @param    subjectId
    *           The subject ID associated with this appserver.
    * @param    sessionFree
    *           Flag indicating if the connection request is for a session-free operating mode.
    * @param    user
    *           The first parameter to pass to the connect procedure, if any.
    * @param    pwd
    *           The second parameter to pass to the connect procedure, if any.
    * @param    serverInfo
    *           The third parameter to pass to the connect procedure, if any.
    *           
    * @return   An {@link AppServerInvocationResult} instance with any result or caught errors.
    */
   AppServerInvocationResult connect(String  subjectId,
                                     boolean sessionFree,
                                     String  user,
                                     String  pwd,
                                     String  serverInfo)
   {
      AppServerInvocationResult result = new AppServerInvocationResult();
   
      AppServerOperatingMode operatingMode = getOperatingMode();
      if (operatingMode.isSessionFree() != sessionFree)
      {
         final String msg =
            "The %s connection failed because the AppServer is in a %s operating mode (%s)";
         final String err = String.format(msg,
                                          sessionFree ? "Session-free" : "Session-managed",
                                          sessionFree ? "session-managed" : "session-free",
                                          operatingMode);
         result.setError(new NumberedException(err, 13511));
         return result;
      }
   
      String id = null;
      try
      {
         id = manager.connect();
      }
      catch (NumberedException e)
      {
         result.setError(e);
         return result;
      }
      
      synchronized(idsToAppserver)
      {
         idsToAppserver.put(id, appDef.appServer);
      }
      
      String connect = appDef.connect;
      if (!operatingMode.isSessionFree() && connect != null && connect.length() > 0)
      {
         Agent agent = acquireAgent(id);
         agent.invoke(result, id, Agent.InvocationModes.AGENT_WAIT, connect, null, "Connect", 
                      "III", new character(user), new character(pwd), new character(serverInfo));
         releaseAgent(id, agent);
         if (result.getError() != null)
         {
            // in case of error, disconnect
            disconnect(id);
   
            id = null;
         }
      }
      
      result.setConnectionID(id);
      
      // regardless of OK/error, the result is sent
      return result;
   }

   /**
    * Disconnect the appserver connection, with the given ID.  It will invoke any configured 
    * disconnect procedure, but will not return any errors or values to the caller.
    * 
    * @param    id
    *           The connection ID to disconnect.
    */
   void disconnect(String id)
   {
      AppServerOperatingMode operatingMode = getOperatingMode();
      
      String disconnect = appDef.disconnect;
      if (!operatingMode.isSessionFree() && disconnect != null && disconnect.length() > 0)
      {
         boolean stateless = (appDef.operatingMode == AppServerOperatingMode.STATELESS);
         AppServerInvocationResult result = new AppServerInvocationResult();
         
         Agent agent = acquireAgent(id);
         // call the disconnect procedure.
         InvocationModes mode = (stateless ? InvocationModes.AGENT_WAIT 
                                           : InvocationModes.AGENT_NO_WAIT);
         agent.invoke(result, id, mode, disconnect, null, "Disconnect", null);
      }
      else
      {
         terminateConnection(id);
      }
   }

   /**
    * Start listening for appserver-related commands.
    * 
    * @param    isAgent
    *           <code>true</code> if this is an agent thread, <code>false</code> if this is the
    *           manager thread.
    */
   void start(boolean isAgent)
   {
      if (isAgent)
      {
         // this is a new Agent for the appServer. do not allow to start more than maxAgents
         int max = appDef.maxAgents;
         if (size() >= max)
         {
            if (LOG.isWarnEnabled())
            {
               LOG.warn("Pool capacity of " + max + " reached, no more agents will be started.");
            }
   
            // just terminate the client, nothing else to do
            return;
         }
         
         // add it to the pool and let it listen for commands
         Agent agent = new Agent(this, appDef.propath);
         poolAgent(agent, true);
         try
         {
            agent.listen(false);
         }
         finally
         {
            // the agent was terminated, remove it from the pool
            removeAgent(agent);
         }
      }
      else
      {
         String subject = SecurityManager.getInstance().getUserId();
         synchronized (subjectToAppserver)
         {
            subjectToAppserver.put(subject, appDef.appServer);
         }
   
         // I'm the agent manager - start the agents
         for (int i = 0; i < appDef.initialAgents; i++)
         {
            // send the command to p2j client side to start the agent.
            // the pool will be populated when the client is started
            handler.startAgent();
         }
         
         // start the trim-agents timer
         final Timer t = new Timer(appDef.appServer + " agent trimming thread.", true);
         TimerTask tt = new TimerTask()
         {
            public void run()
            {
               manager.trimAgents();
            }
         };
         long timeout = appDef.autoTrimTimeout * 1000;
         t.scheduleAtFixedRate(tt, timeout, timeout);
         
         try
         {
            manager.listen(true);
         }
         finally
         {
            // terminate the timer
            t.cancel();
   
            // go through all agents and terminate them
            Set<Agent> agents = null;
            synchronized (all)
            {
               agents = new HashSet<Agent>(all);
            }
   
            for (Agent agent : agents)
            {
               agent.terminate();
            }
         }
      }
   }

   /**
    * Get the manager for this appserver.
    * 
    * @return   See above.
    */
   Agent getManager()
   {
      return manager;
   }

   /**
    * Send the terminate command to all agents in this pool.
    */
   void terminateAll()
   {
      // clean this map first
      synchronized (subjectToAppserver)
      {
         Iterator<String> iter = subjectToAppserver.values().iterator();
         while (iter.hasNext())
         {
            String appServer = iter.next();
            if (appServer.equals(appDef.appServer))
            {
               iter.remove();
            }
         }
      }

      for (Agent agent : all)
      {
         agent.terminate();
      }
   }

   /**
    * Terminate the connection with the given ID.
    * 
    * @param    id
    *           The connection ID.
    */
   void terminateConnection(String id)
   {
      synchronized(idsToAppserver)
      {
         idsToAppserver.remove(id);
      }
   
      Map<String, PersistentProcedure> procs = null;
      synchronized (persistentProcedures)
      {
         procs = persistentProcedures.remove(id);
      }
   
      if (procs != null)
      {
         // delete all resources
         for (PersistentProcedure proc : procs.values())
         {
            proc.delete();
         }
      }
   }

   /**
    * Get the operating mode for this appserver.
    * 
    * @return   See above.
    */
   AppServerOperatingMode getOperatingMode()
   {
      return appDef.operatingMode;
   }

   /**
    * Add an agent back to the {@link #mainPool}.
    * 
    * @param    agent
    *           The agent to add back to the pool.
    * @param    newlyCreated
    *           When this flag is <code>true</code>, the agent is added to the {@link #all global}
    *           pool too.
    */
   void poolAgent(Agent agent, boolean newlyCreated)
   {
      agent.unbind();
      
      synchronized (mainPool)
      {
         mainPool.add(agent);
         mainPool.notify();
      }
   
      if (newlyCreated)
      {
         synchronized (all)
         {
            all.add(agent);
            all.notify();
         }
      }
   }

   /**
    * Remove the given agent entirely from the pool.
    * 
    * @param    agent
    *           The agent which was terminated and needs to be removed.
    */
   void removeAgent(Agent agent)
   {
      synchronized (all)
      {
         all.remove(agent);
         all.notify();
      }
      
      synchronized (mainPool)
      {
         mainPool.remove(agent);
      }
   }

   /**
    * Get the pool size.
    * 
    * @return   See above.
    */
   int size()
   {
      synchronized (all)
      {
         return all.size();
      }
   }

   /**
    * Save the given persistent procedure and associate it with the given connection ID.
    * 
    * @param    id
    *           The connection ID.
    * @param    procId
    *           The procedure's ID.
    * @param    proc
    *           The persistent procedure.
    */
   void saveProcedure(String id, String procId, PersistentProcedure proc)
   {
      Map<String, PersistentProcedure> procs = null;
      synchronized (persistentProcedures)
      {
         procs = persistentProcedures.get(id);
         if (procs == null)
         {
            persistentProcedures.put(id, procs = new HashMap<String, PersistentProcedure>());
         }
      }
      
      procs.put(procId, proc);
   }

   /**
    * Check if the given persistent procedure was started as a remote procedure with the appserver
    * associated with this pool.
    * 
    * @param    proc
    *           The persistent procedure.
    *           
    * @return   See above.
    */
   boolean hasProcedure(PersistentProcedure proc)
   {
      synchronized (persistentProcedures)
      {
         for (Map<String, PersistentProcedure> procs : persistentProcedures.values())
         {
            if (procs.containsValue(proc))
               return true;
         }
      }
      
      return false;
   }
   
   /**
    * Get the persistent procedure associated with the given ID; this returns a non-null instance
    * only if the procedure was not deleted from the context specified by the connection ID.
    *   
    * @param    id
    *           The connection ID.
    * @param    procId
    *           The procedure ID.
    *           
    * @return   The persistent procedure or <code>null</code> if none found.
    */
   PersistentProcedure getProcedure(String id, String procId)
   {
      synchronized (persistentProcedures)
      {
         return persistentProcedures.get(id).get(procId);
      }
   }

   /**
    * Trim the running agents so the pool size is as close to {@link AppServerDefinition#minAgents}
    * as possible.
    */
   void trimAgents()
   {
      final int size = appDef.minAgents;
      synchronized (all)
      {
         int n = all.size() - size;
         // check if we can to terminate any agents
         if (n <= 0)
         {
            return;
         }
   
         for (Agent a : all)
         {
            synchronized (mainPool)
            {
               // if this agent is not busy, terminate it
               if (mainPool.contains(a))
               {
                  // remove it from main pool first, so it can't be picked by other
                  // connections
                  mainPool.remove(a);
                  
                  // terminate
                  a.terminate();
               }
            }
   
            // exit if no more trimming possible
            if (--n == 0)
            {
               break;
            }
         }
         
         all.notify();
      }
   }

   /**
    * Extract an agent from the main pool; if non available, attempt to start one. If the max
    * number of running agents is reached and all agents are busy, return <code>null</code>.
    * 
    * @return   The extracted agent or <code>null</code> if all agents are busy.
    */
   Agent getFromMainPool()
   {
      Agent agent = null;
      long timeout = appDef.requestTimeout * 1000;
      synchronized (mainPool)
      {
         if (mainPool.size() == 0)
         {
            try
            {
               mainPool.wait(timeout);
            }
            catch (InterruptedException e)
            {
               // ignore
            }
         }
   
         if (mainPool.size() > 0)
         {
            agent = mainPool.iterator().next();
            mainPool.remove(agent);
         }
   
      }

      int prevSize = size();
      if (agent == null && prevSize < appDef.maxAgents)
      {
         synchronized (all)
         {
            // automatically start an agent, until maxAgents is reached
            handler.startAgent();
            
            try
            {
               all.wait(timeout);
            }
            catch (InterruptedException e)
            {
               // ignore
            }
         }
         
         // if an agent was started, get it from the main pool.
         if (size() > prevSize)
         {
            agent = getFromMainPool();
         }
      }
      
      return agent;
   }

   /**
    * Class implementing the behavior for STATE-RESET and STATE-AWARE operating modes.  In these
    * modes, an agent is bound to the connection when the connection is established.
    */
   static class BoundPool
   extends AgentPool
   {
      /** The reserved agent, for each connection. */
      private final Map<String, Agent> connections = new HashMap<String, Agent>();

      /** Flag indicating this is STATE-RESET mode. */
      private final boolean reset;
      
      /**
       * Build a new agent pool, for the specified appserver.
       * 
       * @param    appDef
       *           The appserver definition.
       * @param    handler
       *           The handler used to start agents.
       * @param    reset
       *           Flag indicating this is a STATE-RESET mode.
       */
      BoundPool(AppServerDefinition appDef, AppServer handler, boolean reset)
      {
         super(appDef, handler);
         this.reset = reset;
      }
      
      /**
       * Get the agent associated with the given connection id.
       * 
       * @param    id
       *           The connection ID.
       *           
       * @return   See above.
       */
      Agent acquireAgent(String id)
      {
         synchronized (connections)
         {
            return connections.get(id);
         }
      }
      
      /**
       * This is a no-op, as the agent is released only when the connection is terminated.
       * 
       * @param    id
       *           The connection ID.
       * @param    agent
       *           The agent which needs to be released.
       */
      void releaseAgent(String id, Agent agent)
      {
         // no-op, agent is kept bound for the duration of the connection
         
         if (reset)
         {
            // TODO: reset the agent's context
         }
      }
      
      /**
       * Establish a new connection.
       * 
       * @return   The connection ID.
       * 
       * @throws   NumberedException
       *           If the connection could not be established.
       */
      String newConnection()
      throws NumberedException
      {
         String id = nextId();
         // must reserve one agent for this bound connection
         Agent agent = getFromMainPool();
         if (agent == null)
         {
            throw new NumberedException("No servers available");
         }
         
         synchronized (connections)
         {
            connections.put(id, agent);
         }

         agent.bind();
         
         return id;
      }

      /**
       * Terminate the connection with the given ID and unbind the agent.
       * 
       * @param    id
       *           The connection ID.
       */
      void terminateConnection(String id)
      {
         super.terminateConnection(id);
         
         Agent agent = null;
         synchronized (connections)
         {
            agent = connections.remove(id);
         }
         
         poolAgent(agent, false);
      }
   }
   
   /**
    * Class implementing the behavior for STATE-FREE and STATELESS operating modes.  In these
    * modes, no agent is bound to the connection when the connection is established.
    */
   static class UnboundPool
   extends AgentPool
   {
      /**
       * The reserved agent, for each connection.  This map is populated only when the agent
       * becomes bound.
       */
      private final Map<String, Agent> connections = new HashMap<String, Agent>();
      
      /** Flag indicating this is a STATE-FREE mode. */
      private final boolean free;
      
      /**
       * Build a new agent pool, for the specified appserver.
       * 
       * @param    appDef
       *           The appserver definition.
       * @param    handler
       *           The handler used to start agents.
       * @param    free
       *           Flag indicating this is a STATE-FREE mode.
       */
      UnboundPool(AppServerDefinition appDef, AppServer handler, boolean free)
      {
         super(appDef, handler);
         this.free = free;
      }

      /**
       * Get the agent associated with the given connection id.  If no agent is yet bound, it
       * extracts a new agent from the pool.
       * 
       * @param    id
       *           The connection ID.
       *           
       * @return   See above.
       */
      Agent acquireAgent(String id)
      {
         Agent agent = null;
         
         // first check if we have a bound agent
         synchronized (connections)
         {
            agent = connections.get(id);
         }
         
         if (agent == null)
         {
            // no bound agent, get one from the main pool
            agent = getFromMainPool();
         }
         
         return agent;
      }

      /**
       * Release the given agent and place it back in the pool, but only if is still unbound.
       * 
       * @param    id
       *           The connection ID.
       * @param    agent
       *           The agent which needs to be released.
       */
      void releaseAgent(String id, Agent agent)
      {
         if (agent.isBound())
         {
            // agent became bound, save it
            synchronized (connections)
            {
               connections.put(id, agent);
            }
         }
         else
         {
            synchronized (connections)
            {
               connections.remove(id);
            }
            
            // if there was a bound agent, add it back to the main pool
            poolAgent(agent, false);
         }
      }
      
      /**
       * Establish a new connection.
       * 
       * @return   The connection ID.
       * 
       * @throws   NumberedException
       *           If the connection could not be established.
       */
      String newConnection()
      {
         String id = nextId();
      
         synchronized (connections)
         {
            // no agent is initially bound to this pool
            connections.put(id, null);
         }
      
         return id;
      }

      /**
       * Terminate the connection with the given ID and unbind the agent.
       * 
       * @param    id
       *           The connection ID.
       */
      void terminateConnection(String id)
      {
         super.terminateConnection(id);

         Agent agent = null;
         synchronized (connections)
         {
            agent = connections.remove(id);
         }

         if (agent != null)
         {
            // if there was a bound agent, add it back to the main pool
            poolAgent(agent, false);
         }
      }
   }
   
   /**
    * The definition of an appserver.
    */
   private static class AppServerDefinition
   {
      /** The appserver's name. */
      final String appServer;

      /** The appserver's operating mode. */
      final AppServerOperatingMode operatingMode;
      
      /**
       * The amount of time (in seconds) to wait for an agent to be available, before attempting
       * to auto-start a new one. 
       */
      final Integer requestTimeout;
      
      /**
       * The number of seconds waiting before auto-trimming the agents, if they exceed the
       * minimum capacity.
       */
      private final Integer autoTrimTimeout;
      
      /** Appserver's explicit propath. */
      final String propath;
      
      /** The number of agents to start initially. */
      private final Integer initialAgents;
      
      /** The minimum number of agents required for this appserver. */
      final Integer minAgents;
      
      /** The maximum number of agents which can be alive. */
      final Integer maxAgents;
      
      /** The activate procedure; may be null/empty. */
      final String activate;
      
      /** The deactivate procedure; may be null/empty. */
      final String deactivate;

      /** The connect procedure; may be null/empty. */
      private final String connect;

      /** The disconnect procedure; may be null/empty. */
      private final String disconnect;

      /** The startup procedure; may be null/empty. */
      private final String startup;

      /** The shutdown procedure; may be null/empty. */
      private final String shutdown;

      /** The parameter for the startup procedure; may be null/empty. */
      private final String startupParameter;

      /**
       * Read the definition for the given appserver, and save it in this instance.
       * 
       * @param    appServer
       *           The application's server name.
       * 
       * @throws   IllegalArgumentException
       *           If no appserver (with the given name) is configured in the directory.
       */
      AppServerDefinition(String appServer)
      {
         this.appServer = appServer;

         // access the directory service
         DirectoryService ds = DirectoryService.getInstance();
         
         if (!ds.bind())
            throw new RuntimeException("directory bind failed");

         String appSrvPath = Utils.findDirectoryNodePath(ds,
                                                         "appservers/" + appServer,
                                                         "appserver",
                                                         false);

         if (appSrvPath == null)
         {
            throw new IllegalArgumentException(
               "No appserver named '" + appServer +"' is configured in the directory!");
         }
         
         String mode = ds.getNodeString(appSrvPath, "operating_mode");
         this.operatingMode     = AppServerOperatingMode.fromString(mode);
         this.requestTimeout    = ds.getNodeInteger(appSrvPath, "request_timeout");
         this.autoTrimTimeout   = ds.getNodeInteger(appSrvPath, "auto_trim_timeout");

         this.propath          = ds.getNodeString(appSrvPath, "propath");
         this.initialAgents    = ds.getNodeInteger(appSrvPath, "initial_agents");
         this.minAgents        = ds.getNodeInteger(appSrvPath, "min_agents");
         this.maxAgents        = ds.getNodeInteger(appSrvPath, "max_agents");
         this.activate         = ds.getNodeString(appSrvPath, "activate");
         this.deactivate       = ds.getNodeString(appSrvPath, "deactivate");
         this.connect          = ds.getNodeString(appSrvPath, "connect");
         this.disconnect       = ds.getNodeString(appSrvPath, "disconnect");
         if (operatingMode.isSessionFree() && LOG.isWarnEnabled())
         {
            if (LOG.isWarnEnabled() && connect != null && connect.length() > 0)
            {
               LOG.warn("Connect procedure specified for appserver " + appServer + 
                         " in Session-free operating mode");
            }
            if (LOG.isWarnEnabled() && disconnect != null && disconnect.length() > 0)
            {
               LOG.warn("Disconnect procedure specified for appserver " + appServer + 
                        " in Session-free operating mode");
            }
         }
         this.startup          = ds.getNodeString(appSrvPath, "startup");
         this.shutdown         = ds.getNodeString(appSrvPath, "shutdown");
         this.startupParameter = ds.getNodeString(appSrvPath, "startup_parameter");

         // last use of the directory, end our session with it
         ds.unbind();
      }
   }
}
