/*
** Module   : AppServerHandler.java
** Abstract : Implementation for the AppServer interface.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20130529 Created initial version.
*/

package com.goldencode.p2j.util;

import java.io.*;
import java.lang.management.*;
import java.util.*;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.goldencode.p2j.cfg.*;

/**
 * This is the implementation of the network server which allows startup of P2J Clients.
 */
public class AppServerHandler
implements AppServer
{
   /** Logger */
   private static final Log LOG = LogFactory.getLog(AppServerHandler.class);
   
   /** The bootstrap config used for the current authenticated P2J Client. */
   private BootstrapConfig config;

   /** The arguments used to start this P2J Client. */
   private String[] args;

   /**
    * Configure this instance, so that it can start new P2J Clients with the same startup
    * parameters as the running one.
    * 
    * @param    config
    *           The bootstrap config for this running client.
    * @param    args
    *           The arguments used to start this client.
    */
   public AppServerHandler(BootstrapConfig config, String[] args)
   {
      this.config = config;
      this.args = args;
   }
   
   /**
    * Start a new P2J Client, with the same configuration as the running client. This is used
    * by the appserver implementation to start new clients.
    * <p>
    * This inherits, from the current process, the working dir and the environment.
    */
   public void startAgent()
   {
      // this is called on the p2j-client side to start agents
      
      ProcessBuilder pb = new ProcessBuilder();
      // set the working dir to be the same as this dir's working dir
      pb.directory(new File(System.getProperty("user.dir")));
      // environment is automatically set as a copy of this proc's env
      
      // build the command
      List<String> command = buildCommand();
      pb.command(command);
      //pb.inheritIO();
      try
      {
         // no pooling is needed here. if an agent needs to be terminated, just send the TERMINATE
         // command to it
         if (LOG.isInfoEnabled())
         {
            LOG.info("New agent shas been started with command " + command.toString());
         }
         pb.start();
      }
      catch (Throwable t)
      {
         if (LOG.isErrorEnabled())
         {
            LOG.error("Error starting agent with command " + command.toString(), t);
         }
      }
   }

   /**
    * Build the P2J client startup command using, from the current P2J client:
    * <ul>
    *    <li>classpath</li>
    *    <li>library path</li>
    *    <li>JVM arguments.  Any debug-related JVM arguments are dropped</li>
    *    <li>bootstrap config settings</li>
    * <ul>
    * 
    * @return   A list containing the startup command.
    */
   private List<String> buildCommand()
   {
      List<String> command = new ArrayList<String>();
      
      RuntimeMXBean bean = ManagementFactory.getRuntimeMXBean();
      String cp = bean.getClassPath();
      String lp = bean.getLibraryPath();

      command.add("java");
      // add JVM args
      for (String jvmArg : bean.getInputArguments())
      {
         // skip the debug args for now
         // TOOD: compute debug socket
         if (jvmArg.equals("-Xdebug")       ||
             jvmArg.startsWith("-Xrunjdwp") ||
             jvmArg.startsWith("-agentlib"))
            continue;
         
         command.add(jvmArg);
      }
      
      // add library and class paths
      command.add(String.format("-Djava.library.path=%s", lp));
      command.add("-classpath");
      command.add(cp);
      
      // add main class
      command.add("com.goldencode.p2j.main.ClientDriver");
      
      // add cfg args
      String[] categs = config.listCategories();
      final String cfgFormat = "%s:%s:%s=%s";
      for (int i = 0; i < categs.length; i++)
      {
         String categ = categs[i];

         String[] groups = config.listGroups(categ);
         for (int j = 0; j < groups.length; j++)
         {
            String group = groups[j];
            
            String[] keys = config.listKeys(categ, group);
            for (int k = 0; k < keys.length; k++)
            {
               String key = keys[k];
               String val;
               try
               {
                  val = config.getConfigItem(categ, group, key);
               }
               catch (ConfigurationException e)
               {
                  // throw an error here, but this should never happen
                  throw new RuntimeException(e);
               }
               
               String cfgArg = String.format(cfgFormat, categ, group, key, val);
               command.add(cfgArg);
            }
         }
      }
      
      // TODO: what are the app args in this.args?
      
      return command;
   }
}
