/*
** Module   : AppServerOperatingMode.java
** Abstract : Define the available operating-modes for an appserver.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20130529 Created initial version.
*/

package com.goldencode.p2j.util;

/**
 * Defines the appserver's operating modes:
 * <ul>
 *    <li>{@link #STATE_AWARE} for the "State-aware" mode</li>
 *    <li>{@link #STATE_RESET} for the "State-reset" mode</li>
 *    <li>{@link #STATELESS} for the "Stateless" mode</li>
 *    <li>{@link #STATE_FREE} for the "State-free" mode</li>
 * </ul>
 *
 */
enum AppServerOperatingMode
{
   STATE_RESET(false) { public String toString() { return "State-reset"; } },
   STATE_AWARE(false) { public String toString() { return "State-aware"; } },
   STATELESS(false)   { public String toString() { return "Stateless"; } },
   STATE_FREE(true)   { public String toString() { return "State-free"; } };
   
   /**
    * Flag indicating this is a session-free operating mode.
    */
   private final boolean sessionFree;
   
   /**
    * Build a new value and set its {@link #sessionFree} flag.
    * 
    * @param    sessionFree
    *           <code>true</code> if this is a session-free operating mode.
    */
   private AppServerOperatingMode(boolean sessionFree)
   {
      this.sessionFree = sessionFree;
   }
   
   /**
    * Create the appropriate enum constant, from the given string. It must match the constant's
    * {@link #toString()} text.
    * 
    * @param    txt
    *           The string to search, case-insensitively.
    *           
    * @return   The found constant.
    * 
    * @throws   IllegalArgumentException
    *           If the passed text can't be matched with a valid operating mode.
    */
   public static AppServerOperatingMode fromString(String txt)
   {
      for (AppServerOperatingMode mode : AppServerOperatingMode.values())
      {
         if (mode.toString().equalsIgnoreCase(txt))
         {
            return mode;
         }
      }
      
      throw new IllegalArgumentException("Unknown operating mode: '" + txt + "'");
   }

   /**
    * Getter for the {@link #sessionFree} flag.
    * 
    * @return   See above.
    */
   public boolean isSessionFree()
   {
      return sessionFree;
   }
   
   /**
    * Check if this is a session-managed operating mode.
    * 
    * @return   <code>true</code> if this is not a session-free operating mode.
    */
   public boolean isSessionManaged()
   {
      return !sessionFree;
   }
}
